#ifndef DisplayRGB_H_
#define DisplayRGB_H_   291

#define LEFT 0
#define RIGHT 9999
#define CENTER 9998

#define PORTRAIT 0
#define LANDSCAPE 1

//#define USE_SERIAL

//#if ARDUINO < 165
//#define USE_GFX_KBV
//#include "ADA_GFX_kbv.h"
//#else
//#include "Adafruit_GFX.h"
//#endif

//#if ARDUINO >= 100
#include <Arduino.h>
// #include <Print.h>
//#else
// #include "WProgram.h"
//#endif
#include "settings.c"
#include <SPI.h>
#include <avr/pgmspace.h>
#include <petit_fatfs.h>
#include <pff.h>
#include <integer.h>
#include <diskio.h>

// 16 PREDEFINED COLORS UNCOMMENT AND PASTE TO SKETCH
#define BLACK 0x0000 
#define WHITE 0xFFFF 
//#define DARKGRAY 0xAD55 
//#define GRAY 0x8410
//#define BLUE 0x001F 
//#define DARKBLUE 0x0011 
//#define TURQUOISE 0x471A 
//#define GREEN 0x0400 
//#define FORESTGREEN 0x2444 
//#define YELLOW 0xFFE0 
//#define ORANGE 0xFD20 
//#define PINK 0xFE19 
//#define RED 0xF800 
//#define PURPLE 0x8010 
//#define VIOLET 0xEC1D 
//#define BROWN 0xA145
//#define BEIGE 0xF7BB

//for extended bitmap function :
#define NOROT 0
#define ROTCCW 1
#define ROT180 2
#define ROTCW 3
#define NOFLIP 0
#define FLIPH 1
#define FLIPV 2
#define FLIPVH 3

#if (DISPLAY_ROT == NO_ROT)||(DISPLAY_ROT == ROT180) //landscape mode
#define LCDWIDTH 320
#define LCDHEIGHT 240
#else //portrait mode
#define LCDWIDTH 240
#define LCDHEIGHT 320
#endif
#define LCDHEIGHT_NOROT 240
#define LCDWIDTH_NOROT 320

#define swap(a, b) { int16_t t = a; a = b; b = t; }

extern uint8_t _displayBuffer[];


class Display : public Print {

public:
	//#if defined USE_GFX_KBV
	//	MCUFRIEND_kbv();
	//#else
	void tftBegin(int CS = A3, int RS = A2, int WR = A1, int RD = A0, int RST = A4);
	//#endif

	void     reset(void);                                       // you only need the constructor
	void     begin(uint16_t ID = 0x9341);                       // you only need the constructor
	virtual void     drawPixel(int16_t x, int16_t y, uint16_t color);  // and these three
	void drawPixelPC(int x, int y) { Display::drawPixel(x, y, _fcolor); }
	void     WriteCmdData(uint16_t cmd, uint16_t dat);                 // ?public methods !!!
	void     pushCommand(uint16_t cmd, uint8_t * block, int8_t N);

	void clrScrPC() { Display::fillScreen(0x0000); }
	void printPC(const char *st, int x, int y, int deg = 0) { settextcursor((char*)st, x, y); Display::print(st); }
	void printPC(char *st, int x, int y, int deg = 0) { settextcursor(st, x, y); Display::print(st); }
	void printPC(String st, int x, int y, int deg = 0) { Display::print(st); }
	void printNumIPC(long num, int x, int y, int length = 0, char filler = ' ') { char buf[16]; ltoa(num, buf, 10); settextcursor(buf, x, y); Display::print(buf); }
	void printNumFPC(double num, byte dec, int x, int y, char divider = '.', int length = 0, char filler = ' ') { settextcursor((char*)"", x, y); Display::print(num, dec); }

	uint8_t* getBuffer(); //duhjoker from gamebuino
	uint16_t color565(uint8_t r, uint8_t g, uint8_t b) { return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | ((b & 0xF8) >> 3); }
	uint16_t readID(void);

	void clear(void);

	virtual void     drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1, uint16_t color);
	        void     drawLinePC(int x1, int y1, int x2, int y2) { Display::drawLine(x1, y1, x2, y2, _fcolor); }
	virtual void     drawFastVLine(int16_t x, int16_t y, int16_t h, uint16_t color);
	virtual void     drawFastHLine(int16_t x, int16_t y, int16_t w, uint16_t color);

	virtual void     drawRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
	        void     drawRectPC(int x1, int y1, int x2, int y2) { int w = x2 - x1 + 1, h = y2 - y1 + 1; if (w < 0) { x1 = x2; w = -w; }if (h < 0) { y1 = y2; h = -h; } Display::drawRect(x1, y1, w, h, _fcolor); }
	virtual void     fillRect(int16_t x, int16_t y, int16_t w, int16_t h, uint16_t color);
        	void     fillRectPC(int x1, int y1, int x2, int y2) { int w = x2 - x1 + 1, h = y2 - y1 + 1; if (w < 0) { x1 = x2; w = -w; } if (h < 0) { y1 = y2; h = -h; } Display::fillRect(x1, y1, w, h, _fcolor); } 
	virtual void     fillScreen(uint16_t color);
	        void     fillScrPC(byte r, byte g, byte b) { Display::fillScreen(setrgb(r, g, b)); }

	void drawCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color);
	void drawCirclePC(int x, int y, int radius) { Display::drawCircle(x, y, radius, _fcolor); }
	void drawCircleHelper(int16_t x0, int16_t y0, int16_t r, uint8_t cornername, uint16_t color);
	void fillCircle(int16_t x0, int16_t y0, int16_t r, uint16_t color);
	void fillCirclePC(int x, int y, int radius) { Display::fillCircle(x, y, radius, _fcolor); }
	void fillCircleHelper(int16_t x0, int16_t y0, int16_t r, uint8_t cornername, int16_t delta, uint16_t color);

	void drawTriangle(int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color);
	void fillTriangle(int16_t x0, int16_t y0, int16_t x1, int16_t y1, int16_t x2, int16_t y2, uint16_t color);

	void drawRoundRect(int16_t x0, int16_t y0, int16_t w, int16_t h, int16_t radius, uint16_t color);
	void drawRoundRectPC(int x1, int y1, int x2, int y2) { int w = x2 - x1 + 1, h = y2 - y1 + 1; if (w < 0) { x1 = x2; w = -w; }if (h < 0) { y1 = y2; h = -h; } Display::drawRoundRect(x1, y1, w, h, _radius, _fcolor); }
	void fillRoundRect(int16_t x0, int16_t y0, int16_t w, int16_t h, int16_t radius, uint16_t color);
	void fillRoundRectPC(int x1, int y1, int x2, int y2) { int w = x2 - x1 + 1, h = y2 - y1 + 1; if (w < 0) { x1 = x2; w = -w; } if (h < 0) { y1 = y2; h = -h; } Display::fillRoundRect(x1, y1, w, h, _radius, _fcolor); }
	
	//adafruit_gfx 
	void drawBitmap1(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color);
	void drawBitmap2(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, uint16_t bg);
	void drawBitmap3(int16_t x, int16_t y, uint8_t *bitmap, int16_t w, int16_t h, uint16_t color);
	void drawBitmap4(int16_t x, int16_t y, uint8_t *bitmap, int16_t w, int16_t h, uint16_t color, uint16_t bg);
	void drawXBitmap(int16_t x, int16_t y, const uint8_t *bitmap, int16_t w, int16_t h, uint16_t color);

	//GAMEBUINO----SUMMONER123
	void drawBitmap(int8_t x, int8_t y, int8_t w, int8_t h, const uint8_t *bitmap, uint8_t dx, uint8_t dy, uint8_t dw, uint8_t dh, uint16_t color);
	boolean getBitmapPixel(const uint8_t* bitmap, uint8_t x, uint8_t y);

	//draw from SD  mouguino    
	//Draw bitmap from SD at position x0, y0
	void drawBitmapFromSdA(byte x, byte y, char *bitmap, uint16_t color);
	void drawBitmapFromSdB(byte x, byte y, char *bitmap, uint8_t dx, uint8_t dy, uint8_t dw, uint8_t dh, uint16_t color);
	int drawBitmapFromSd(byte x0, byte y0, char *bitmap, uint16_t color);

	//void updateSlide();

	//GAMEBUINO----SUMMONER123
	void drawTilemap(int x, int y, const uint8_t *tilemap, const uint8_t **spritesheet, uint16_t color);
	void drawTilemap(int x, int y, const uint8_t *tilemap, const uint8_t **spritesheet, uint8_t dx, uint8_t dy, uint8_t dw, uint8_t dh, uint16_t color);

	//Drawbitmap pushcolors
	void drawBitmapPC(int x, int y, int sx, int sy, const uint16_t *data, int scale = 1) {
		uint16_t color;
		Display::setAddrWindow(x, y, x + sx*scale - 1, y + sy*scale - 1);
		if (scale == 1) Display::pushColors((const uint8_t*)data, sx * sy, 1);
		else {
			for (int row = 0; row < sy; row++) {
				for (int col = 0; col < sx; col++) {
					color = pgm_read_word(data + (row*sx + col) * 1);
					Display::fillRect(x + col*scale, y + row*scale, scale, scale, color);
				}
			}
		}
	}

	virtual void     invertDisplay(boolean i);

	uint16_t readReg(uint16_t reg);
	uint32_t readReg32(uint16_t reg);
	int16_t  readGRAM(int16_t x, int16_t y, uint16_t *block, int16_t w, int16_t h);
	uint16_t readPixel(int16_t x, int16_t y) { uint16_t color; readGRAM(x, y, &color, 1, 1); return color; }

	void     setAddrWindow(int16_t x, int16_t y, int16_t x1, int16_t y1);
	void     pushColors(uint16_t *block, int16_t n, bool first);
	void     pushColors(uint8_t *block, int16_t n, bool first);
	void     pushColors(const uint8_t *block, int16_t n, bool first);
	void     vertScroll(int16_t top, int16_t scrollines, int16_t offset);

	void setFont(const uint8_t* f); //duhjoker from gamebuino 
	uint8_t fontWidth, fontHeight;  //duhjoker from gamebuino 

	void drawChar(int16_t x, int16_t y, unsigned char c, uint16_t color, uint16_t bg, uint8_t size);
	void setCursor(int16_t x, int16_t y);
	void setColor(uint16_t color);                //DUHJOKER COLOR LCD SPECIFIC.... USE BEFORE DRAWBITMAP AND TILEMAP FUNCTIONS
//	void setColor(uint16_t color , uint16_t color);   //DUHJOKER COLOR LCD SPECIFIC.... USE BEFORE DRAWBITMAP AND TILEMAP FUNCTIONS
	void setTextColor(uint16_t c);
	void setTextColor(uint16_t c, uint16_t bg);
	void setTextSize(uint8_t s);
	void setTextWrap(boolean w);
	void setRotation(uint8_t r);

#if ARDUINO >= 100
	virtual size_t write(uint8_t);
#else
	virtual void   write(uint8_t);
#endif

	boolean persistence; //disable clean() at each frame if true
	boolean textWrap; // If set, 'wrap' text at right edge of 
	uint8_t fontSize;
	int8_t cursorX, cursorY;
	byte contrast;
	byte frameCount;

	typedef struct {       //line 171 "Public Variables   - ADD by Summoner123
		int x;                    //X coordinate                 - ADD by Summoner123
		int y;                    //Y coordinate                 - ADD by Summoner123
		const byte *spritecol;    //Sprite of object             - ADD b Summoner123
	}object;
	object solid[60];         // Matriz were saved a Sprite, X and Y cordinates of all tiles on the screen - ADD by Summoner123

	byte numcolision = 0;     //count of solid objects indacat how many tiles drawed on the screen - ADD by Summoner123

	bool flagcollision = true;

	int16_t height(void);
	int16_t width(void);

	uint8_t getRotation(void);

protected:
	uint16_t _fcolor;
	uint16_t _bcolor;
	uint8_t _radius;
	uint8_t _orient;
	void settextcursor(char *st, int x, int y) {
		int16_t pos, x1, y1;
		uint16_t len, w, h;
#if defined(_GFXFONT_H_)
		if (Display::gfxFont != NULL) y += pgm_read_byte(&gfxFont->yAdvance);
		if (x == CENTER || x == RIGHT) {
			if (Display::gfxFont != NULL) {
				getTextBounds(st, 0, 0, &x1, &y1, &w, &h);
				len = x1 - 0 + w;    // misses the final right padding.
			}
			else
#else
		if (x == CENTER || x == RIGHT) {
#endif
			{
				len = strlen(st) * 6;
			}
			pos = (Display::width() - len);
			if (x == CENTER) x = pos / 2;
			else x = pos;
		}
		Display::setCursor(x, y);
	}
	uint16_t setrgb(byte r, byte g, byte b) { return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3); }

	int16_t  WIDTH, HEIGHT;   // this is the 'raw' display w/h - never changes
	int16_t  _width, _height; // dependent on rotation
	int16_t  cursor_x, cursor_y;
	uint16_t textcolor, textbgcolor;
	uint8_t  textsize;
	uint8_t  rotation;
	boolean  wrap; // If set, 'wrap' text at right edge of display
	bool useFill;

	private:
	uint16_t _lcd_ID, _lcd_rev, _lcd_madctl, _lcd_drivOut, _MC, _MP, _MW, _SC, _EC, _SP, _EP;
    uint16_t  _lcd_xor, _lcd_capable;

	void set_window(uint8_t x1, uint8_t y1, uint8_t x2, uint8_t y2);
	void write_wr_reg(uint16_t data);
	void write_wr_data(uint16_t data);
	void write_cmd(uint16_t cmd, uint16_t data);
#if 0 //defined(ARDUINO_ARCH_SAM)
	void write_delay();
#else
#define write_delay()  
#endif

	//value	
	//private:

#if defined(ARDUINO_ARCH_SAM)
	uint32_t _cs;
	uint32_t _pmw;
	uint32_t _wr;
	uint32_t _rs;
	uint32_t _rest;

	volatile uint32_t *csport;
	volatile uint32_t *pmwport;
	volatile uint32_t *wrport;
	volatile uint32_t *rsport;
	volatile uint32_t *restport;

	uint32_t cspinmask;
	uint32_t pmwpinmask;
	uint32_t wrpinmask;
	uint32_t rspinmask;
	uint32_t restpinmask;

	volatile uint32_t data[8];
	volatile uint32_t datapinmask[8];
	volatile uint32_t* dataport[8];

#else
	uint8_t _cs;
	uint8_t _pmw;
	uint8_t _wr;
	uint8_t _rs;
	uint8_t _rest;

	volatile uint8_t *csport;
	volatile uint8_t *pmwport;
	volatile uint8_t *wrport;
	volatile uint8_t *rsport;
	volatile uint8_t *restport;

	uint8_t cspinmask;
	uint8_t pmwpinmask;
	uint8_t wrpinmask;
	uint8_t rspinmask;
	uint8_t restpinmask;

	volatile uint8_t data[8];
	volatile uint8_t datapinmask[8];
	volatile uint8_t* dataport[8];

#endif

	uint8_t *font;
};

inline uint8_t* Display::getBuffer() {
	return _displayBuffer;
}

#if ARDUINO >= 165
#include <SPI.h>
#endif

#if 0
#elif defined(__AVR_ATmega328P__)

#define SPI_INIT()   { DDRB |= (1<<5)|(1<<3)|(1<<2); SPCR = (1<<SPE)|(1<<MSTR); SPSR = (1<<SPI2X); SPSR; SPDR; }
static inline uint8_t spi_xfer(uint8_t c)
{
	SPDR = c;
	while ((SPSR & (1 << SPIF)) == 0);
	return SPDR;
}
extern uint8_t running;
static inline void write8(uint8_t x) {
	if (running) {
		while ((SPSR & 0x80) == 0);
		SPDR;
	}
	SPDR = x;
	running = 1;
}
static inline uint8_t read8(void) {
	if (running) while ((SPSR & 0x80) == 0);
	running = 0;
	return SPDR;
}
static inline uint8_t xchg8(uint8_t x) { write8(x); return read8(); }
static inline void flush(void) {
	if (running) {
		while ((SPSR & 0x80) == 0);
	}
	running = 0;
	SPDR;
}

#if defined(SUPPORT_8347D)
#warning using HX8347D hardware
#define CD_PORT PORTD
#define CD_PIN  PD7
#define CS_PORT PORTB
#define CS_PIN  PB2
#define RESET_PORT PORTB
#define RESET_PIN  PB1
#define SD_PIN  PD5
#define XPT_PIN PD4
#define RD_IDLE
#define WR_IDLE
#else
#warning using regular SPI hardware
#define CD_PORT PORTB
#define CD_PIN  1
#define CS_PORT PORTB
#define CS_PIN  2
#define RESET_PORT PORTB
#define RESET_PIN  0
#define RD_IDLE
#define WR_IDLE
#endif

#define setWriteDir() { }
#define setReadDir()  { }
//#define write8(x)     spi_xfer(x)
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { dst = xchg8(0); }
#define READ_16(dst)  { dst = xchg8(0); dst = (dst << 8) | xchg8(0);  }

#define PIN_LOW(p, b)        (p) &= ~(1<<(b))
#define PIN_HIGH(p, b)       (p) |= (1<<(b))
#define PIN_OUTPUT(p, b)     *(&p-1) |= (1<<(b))
#elif defined(__SAMD21G18A__)

#define SPI_INIT()   { SPI.begin(); SPI.setDataMode(SPI_MODE0); SPI.setClockDivider(6); }

#define CD_PORT PORT->Group[0]
#define CD_PIN  7
#define CS_PORT PORT->Group[0]
#define CS_PIN  18
#define RESET_PORT PORT->Group[0]
#define RESET_PIN  6
#define RD_IDLE
#define WR_IDLE


uint8_t running;
static inline void write8(uint8_t c)
{
	running = 1;
	while (SERCOM1->SPI.INTFLAG.bit.DRE == 0);
	SERCOM1->SPI.DATA.bit.DATA = c; // Writing data into Data register
}

static inline void flush(void)
{
	if (running) while (SERCOM1->SPI.INTFLAG.bit.TXC == 0);
	running = 0;
}

static inline uint8_t xchg8(uint8_t c)
{
	//  flush();
	while (SERCOM1->SPI.INTFLAG.bit.RXC != 0) SERCOM1->SPI.DATA.bit.DATA; //eat up
	while (SERCOM1->SPI.INTFLAG.bit.DRE == 0);
	SERCOM1->SPI.DATA.bit.DATA = c; // Writing data into Data register
	while (SERCOM1->SPI.INTFLAG.bit.RXC == 0);
	return SERCOM1->SPI.DATA.bit.DATA;
}


#define setWriteDir() { }
#define setReadDir()  { }
//#define flush()
//#define write8(x) xchg8(x)
//#define xchg8(x)     SPI.transfer(x)
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { dst = xchg8(0); }
#define READ_16(dst)  { dst = xchg8(0); dst = (dst << 8) | xchg8(0);  }

// Shield Control macros.
#define PIN_LOW(port, pin)    (port).OUTCLR.reg = (1<<(pin))
#define PIN_HIGH(port, pin)   (port).OUTSET.reg = (1<<(pin))
#define PIN_OUTPUT(port, pin) (port).DIR.reg |= (1<<(pin))

#elif defined(__AVR_ATxmega128A1__)     //3.49s @ 32MHz -O2
#define CD_PORT VPORT2
#define CD_PIN  1
#define CS_PORT VPORT3
#define CS_PIN  4
#define RESET_PORT VPORT2
#define RESET_PIN  0
#define SPCRVAL (USART_CLK2X_bm | USART_RXEN_bm | USART_TXEN_bm)
#define SETDDR  {VPORT3.DIR |= (1<<4)|(1<<5)|(1<<7); VPORT2.DIR |= 0x03; }
#define SPI_INIT()  { PORTCFG.VPCTRLB=PORTCFG_VP3MAP_PORTF_gc | PORTCFG_VP2MAP_PORTC_gc; CS_IDLE; RESET_IDLE; SETDDR; spi_init(); }

void spi_init(void)
{
	SPIF.CTRL = SPI_ENABLE_bm | SPI_MODE_3_gc | (1 << SPI_MASTER_bp) | (1 << SPI_CLK2X_bp);
}

#define write8(x)    {\
						 SPIF.DATA=x;\
                         while ((SPIF.STATUS & SPI_IF_bm)==0);\
                         SPIF.DATA;\
                     }
#define flush()   {\
                  }

#define PIN_LOW(p, b)        (p).OUT &= ~(1<<(b))
#define PIN_HIGH(p, b)       (p).OUT |= (1<<(b))
#define PIN_OUTPUT(p, b)     (p).DIR |= (1<<(b))

#elif defined(__AVR_ATxmega32A4U__)     //3.49s @ 32MHz -O2.  
// 100ns/150ns for ILI9341 W/R cycle.   100ns/200ns for ILI920.  20ns/150ns HX8347
// Xmega @ 60MHz i.e. 30MHz SCK works with 9341.
#warning Using ATxmega32A4U USART_MSPI
#define CD_PORT VPORT2
#define CD_PIN  1
#define CS_PORT VPORT3
#define CS_PIN  0
#define RESET_PORT VPORT2
#define RESET_PIN  0
#define SD_PORT    PORTC
#define SD_PIN     4
#define SPCRVAL (USART_CLK2X_bm | USART_RXEN_bm | USART_TXEN_bm)
#define SETDDR  {PORTCFG.VPCTRLB=PORTCFG_VP13MAP_PORTD_gc | PORTCFG_VP02MAP_PORTC_gc; VPORT3.DIR |= (1<<0)|(1<<1)|(1<<3); VPORT2.DIR |= 0x03; PIN_HIGH(SD_PORT, SD_PIN); SD_PORT.DIR |= (1<<SD_PIN); }
#define SPI_INIT()  { CS_IDLE; RESET_IDLE; SETDDR; spi_init(); }

static inline void spi_init(void)
{
	USARTD0.CTRLB = SPCRVAL;
	USARTD0.CTRLC = USART_CMODE_MSPI_gc | 0x00 | 0x00;   //mode #0 
														 //   PORTD.PIN1CTRL |= PORT_INVEN_bm;   //CPOL
	USARTD0.BAUDCTRLA = 0x00;     //F_CPU/2
	USARTD0.BAUDCTRLB = ((0x00 << USART_BSCALE_gp) & USART_BSCALE_gm) | 0x00;
	USARTD0.DATA;
}

extern uint8_t running;

#define write8(x)    {\
						 while ((USARTD0.STATUS & USART_DREIF_bm) == 0) ;\
                         asm("cli");\
						 USARTD0.DATA = x;\
						 USARTD0.STATUS = USART_TXCIF_bm;\
						 asm("sei");\
						 running = 1;\
                     }
static inline uint8_t read8(void) {
	if (running) while ((USARTD0.STATUS & USART_RXCIF_bm) == 0);
	return USARTD0.DATA;
}
#define flush()   {\
                         if (running) while ((USARTD0.STATUS & USART_TXCIF_bm) == 0) ;\
						 while ((USARTD0.STATUS & USART_RXCIF_bm) != 0) USARTD0.DATA;\
						 running = 0;\
                  }
static inline uint8_t xchg8(uint8_t x) {
	USARTD0.DATA = x;
	while ((USARTD0.STATUS & USART_RXCIF_bm) == 0);
	return USARTD0.DATA;
}
/*
#define write8(x)    {\
while ((USARTD0.STATUS & USART_DREIF_bm) == 0) ;\
USARTD0.DATA = x;\
while ((USARTD0.STATUS & USART_RXCIF_bm) == 0) ;\
USARTD0.DATA;\
}
#define flush()
*/

#define RD_IDLE
#define WR_IDLE
//#define SPI_INIT()    spi_init()
#define setWriteDir() { }
#define setReadDir()  { }
//#define write8(x)     spi_xfer(x)
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { dst = xchg8(0); }
#define READ_16(dst)  { dst = xchg8(0); dst = (dst << 8) | xchg8(0);  }

#define PIN_LOW(p, b)        (p).OUT &= ~(1<<(b))
#define PIN_HIGH(p, b)       (p).OUT |= (1<<(b))
#define PIN_OUTPUT(p, b)     (p).DIR |= (1<<(b))

#endif

#define CD_COMMAND {flush(); PIN_LOW(CD_PORT, CD_PIN); }
#define CD_DATA    {flush(); PIN_HIGH(CD_PORT, CD_PIN); }
#define CD_OUTPUT  PIN_OUTPUT(CD_PORT, CD_PIN)
#define CS_ACTIVE  PIN_LOW(CS_PORT, CS_PIN)
#define CS_IDLE    {flush(); PIN_HIGH(CS_PORT, CS_PIN); }
#define CS_OUTPUT  PIN_OUTPUT(CS_PORT, CS_PIN)
#define RESET_ACTIVE  PIN_LOW(RESET_PORT, RESET_PIN)
#define RESET_IDLE    PIN_HIGH(RESET_PORT, RESET_PIN)
#define RESET_OUTPUT  PIN_OUTPUT(RESET_PORT, RESET_PIN)

// General macros.   IOCLR registers are 1 cycle when optimised.

#define CTL_INIT()   { CD_OUTPUT; CS_OUTPUT; RESET_OUTPUT; SPI_INIT(); }
#define WriteCmd(x)  { CD_COMMAND; write8(x); }
#define WriteData(x) { CD_DATA; write16(x); }

//#define USE_SPECIAL             //check for custom drivers
#if defined(USE_SPECIAL)
#include "mcufriend_special.h"
#if !defined(USE_SPECIAL_FAIL)
#warning WE ARE USING A SPECIAL CUSTOM DRIVER
#endif
#endif
#if !defined(USE_SPECIAL) || defined (USE_SPECIAL_FAIL)

#if 0
#elif defined(__AVR_ATmega328P__)       //regular UNO shield on UNO
#define RD_PORT PORTC
#define RD_PIN  0
#define WR_PORT PORTC
#define WR_PIN  1
#define CD_PORT PORTC
#define CD_PIN  2
#define CS_PORT PORTC
#define CS_PIN  3
#define RESET_PORT PORTC
#define RESET_PIN  4

#define DMASK         0x03
#define NMASK         ~DMASK
#define write_8(x)    { PORTB = (PORTB & NMASK) | ((x) & DMASK); PORTD = (PORTD & DMASK) | ((x) & NMASK); }
#define read_8()      ( (PINB & DMASK) | (PIND & NMASK) )
#define setWriteDir() { DDRB = (DDRB & NMASK) | DMASK; DDRD = (DDRD & DMASK) | NMASK;  }
#define setReadDir()  { DDRB = (DDRB & NMASK) & NMASK; DDRD = (DDRD & DMASK) & DMASK;  }
#define write8(x)     { write_8(x); WR_STROBE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; dst = read_8(); RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }

#define PIN_LOW(p, b)        (p) &= ~(1<<(b))
#define PIN_HIGH(p, b)       (p) |= (1<<(b))
#define PIN_OUTPUT(p, b)     *(&p-1) |= (1<<(b))

#elif defined(__AVR_ATmega2560__) || defined(__AVR_ATmega1280__)       //regular UNO shield on MEGA2560
#define RD_PORT PORTF
#define RD_PIN  0
#define WR_PORT PORTF
#define WR_PIN  1
#define CD_PORT PORTF
#define CD_PIN  2
#define CS_PORT PORTF
#define CS_PIN  3
#define RESET_PORT PORTF
#define RESET_PIN  4

#define EMASK         0x38
#define GMASK         0x20
#define HMASK         0x78
#define write_8(x)   {  PORTH &= ~HMASK; PORTG &= ~GMASK; PORTE &= ~EMASK; \
                        PORTH |= (((x) & (3<<0)) << 5); \
                        PORTE |= (((x) & (3<<2)) << 2); \
                        PORTG |= (((x) & (1<<4)) << 1); \
                        PORTE |= (((x) & (1<<5)) >> 2); \
                        PORTH |= (((x) & (3<<6)) >> 3); \
					 }

#define read_8()      ( ((PINH & (3<<5)) >> 5)\
                      | ((PINE & (3<<4)) >> 2)\
                      | ((PING & (1<<5)) >> 1)\
                      | ((PINE & (1<<3)) << 2)\
                      | ((PINH & (3<<3)) << 3)\
                      )
#define setWriteDir() { DDRH |=  HMASK; DDRG |=  GMASK; DDRE |=  EMASK;  }
#define setReadDir()  { DDRH &= ~HMASK; DDRG &= ~GMASK; DDRE &= ~EMASK;  }
#define write8(x)     { write_8(x); WR_STROBE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; dst = read_8(); RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }

#define PIN_LOW(p, b)        (p) &= ~(1<<(b))
#define PIN_HIGH(p, b)       (p) |= (1<<(b))
#define PIN_OUTPUT(p, b)     *(&p-1) |= (1<<(b))

#elif defined(__SAMD21G18A__)   //regular UNO shield on ZERO or M0_PRO
#include "sam.h"
// configure macros for the control pins
#define RD_PORT PORT->Group[0]
#define RD_PIN  2
#define WR_PORT PORT->Group[1]
#define WR_PIN  8
#define CD_PORT PORT->Group[1]
#define CD_PIN  9
#define CS_PORT PORT->Group[0]
#define CS_PIN  4
#define RESET_PORT PORT->Group[0]
#define RESET_PIN  5
// configure macros for data bus
#define DMASK 0x0030C3C0
//  #define write_8(x) PORT->Group[0].OUT.reg = (PORT->Group[0].OUT.reg & ~DMASK)|(((x) & 0x0F) << 6)|(((x) & 0x30) << 10)|(((x) & 0xC0)<<14)
#if defined(ARDUINO_SAMD_ZERO) || defined(ARDUINO_SAMD_ZERO)   // American ZERO
#define write_8(x) {\
	PORT->Group[0].OUTCLR.reg = DMASK;\
	PORT->Group[0].OUTSET.reg = (((x) & 0x0B) << 6)\
                               |(((x) & (1<<2)) << 12)\
	                           |(((x) & (1<<4)) << 4)\
	                           |(((x) & (1<<5)) << 10)\
	                           |(((x) & 0xC0) << 14);\
                   }
#define read_8()   (((PORT->Group[0].IN.reg >> 6) & 0x0B)\
                   |((PORT->Group[0].IN.reg >> 12) & (1<<2))\
                   |((PORT->Group[0].IN.reg >> 4) &  (1<<4))\
                   |((PORT->Group[0].IN.reg >> 10) & (1<<5))\
                   |((PORT->Group[0].IN.reg >> 14) & 0xC0))
#else   //default to an M0_PRO on v1.6.5 or 1.7.6
#define write_8(x) {\
	PORT->Group[0].OUTCLR.reg = DMASK;\
	PORT->Group[0].OUTSET.reg = (((x) & 0x0F) << 6)\
                               |(((x) & 0x30) << 10)\
                               |(((x) & 0xC0) << 14);\
                   }
#define read_8()   (((PORT->Group[0].IN.reg >> 6) & 0x0F)|((PORT->Group[0].IN.reg >> 10) & 0x30)|((PORT->Group[0].IN.reg >> 14) & 0xC0))
#endif
#define setWriteDir() { PORT->Group[0].DIRSET.reg = DMASK; \
	                  PORT->Group[0].WRCONFIG.reg = (DMASK & 0xFFFF) | (0<<22) | (1<<28) | (1<<30); \
	                  PORT->Group[0].WRCONFIG.reg = (DMASK>>16) | (0<<22) | (1<<28) | (1<<30) | (1<<31); \
                        }
#define setReadDir()  { PORT->Group[0].DIRCLR.reg = DMASK; \
	                  PORT->Group[0].WRCONFIG.reg = (DMASK & 0xFFFF) | (1<<17) | (1<<28) | (1<<30); \
	                  PORT->Group[0].WRCONFIG.reg = (DMASK>>16) | (1<<17) | (1<<28) | (1<<30) | (1<<31); \
                        }
#define write8(x)     { write_8(x); WR_ACTIVE; WR_STROBE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; dst = read_8(); RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }
// Shield Control macros.
#define PIN_LOW(port, pin)    (port).OUTCLR.reg = (1<<(pin))
#define PIN_HIGH(port, pin)   (port).OUTSET.reg = (1<<(pin))
#define PIN_OUTPUT(port, pin) (port).DIR.reg |= (1<<(pin))

#elif defined(__SAM3X8E__)      //regular UNO shield on DUE
// configure macros for the control pins
#define RD_PORT PIOA
#define RD_PIN  16
#define WR_PORT PIOA
#define WR_PIN  24
#define CD_PORT PIOA
#define CD_PIN  23
#define CS_PORT PIOA
#define CS_PIN  22
#define RESET_PORT PIOA
#define RESET_PIN  6
// configure macros for data bus
#define BMASK         (1<<25)
#define CMASK         (0xBF << 21)
#define write_8(x)   {  PIOB->PIO_CODR = BMASK; PIOC->PIO_CODR = CMASK; \
                        PIOC->PIO_SODR = (((x) & (1<<0)) << 22); \
                        PIOC->PIO_SODR = (((x) & (1<<1)) << 20); \
                        PIOB->PIO_SODR = (((x) & (1<<2)) << 23); \
                        PIOC->PIO_SODR = (((x) & (1<<3)) << 25); \
                        PIOC->PIO_SODR = (((x) & (1<<4)) << 22); \
                        PIOC->PIO_SODR = (((x) & (1<<5)) << 20); \
                        PIOC->PIO_SODR = (((x) & (1<<6)) << 18); \
                        PIOC->PIO_SODR = (((x) & (1<<7)) << 16); \
					 }

#define read_8()      ( ((PIOC->PIO_PDSR & (1<<22)) >> 22)\
                      | ((PIOC->PIO_PDSR & (1<<21)) >> 20)\
                      | ((PIOB->PIO_PDSR & (1<<25)) >> 23)\
                      | ((PIOC->PIO_PDSR & (1<<28)) >> 25)\
                      | ((PIOC->PIO_PDSR & (1<<26)) >> 22)\
                      | ((PIOC->PIO_PDSR & (1<<25)) >> 20)\
                      | ((PIOC->PIO_PDSR & (1<<24)) >> 18)\
                      | ((PIOC->PIO_PDSR & (1<<23)) >> 16)\
                      )
#define setWriteDir() { PIOB->PIO_OER = BMASK; PIOC->PIO_OER = CMASK; }
#define setReadDir()  { \
                          PMC->PMC_PCER0 = (1 << ID_PIOB)|(1 << ID_PIOC);\
						  PIOB->PIO_ODR = BMASK; PIOC->PIO_ODR = CMASK;\
						}
#define write8(x)     { write_8(x); WR_ACTIVE; WR_STROBE; }
//#define write8(x)     { write_8(x); WR_ACTIVE; WR_STROBE; WR_IDLE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; RD_ACTIVE; dst = read_8(); RD_IDLE; RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }
// Shield Control macros.
#define PIN_LOW(port, pin)    (port)->PIO_CODR = (1<<(pin))
#define PIN_HIGH(port, pin)   (port)->PIO_SODR = (1<<(pin))
#define PIN_OUTPUT(port, pin) (port)->PIO_OER = (1<<(pin))

#elif defined(__AVR_ATmega32U4__)       //regular UNO shield on Leonardo
#define RD_PORT PORTF
#define RD_PIN  7
#define WR_PORT PORTF
#define WR_PIN  6
#define CD_PORT PORTF
#define CD_PIN  5
#define CS_PORT PORTF
#define CS_PIN  4
#define RESET_PORT PORTF
#define RESET_PIN  1

#define BMASK         (3<<4)
#define CMASK         (1<<6)
#define DMASK         ((1<<7)|(1<<4)|(3<<0))
#define EMASK         (1<<6)
static inline                   //hope we use r24
void write_8(uint8_t x)
{
	PORTB &= ~BMASK;
	PORTC &= ~CMASK;
	PORTD &= ~DMASK;
	PORTE &= ~EMASK;
	PORTB |= (((x) & (3 << 0)) << 4);
	PORTD |= (((x) & (1 << 2)) >> 1);
	PORTD |= (((x) & (1 << 3)) >> 3);
	PORTD |= (((x) & (1 << 4)) << 0);
	PORTC |= (((x) & (1 << 5)) << 1);
	PORTD |= (((x) & (1 << 6)) << 1);
	PORTE |= (((x) & (1 << 7)) >> 1);
}

#define read_8()      ( ((PINB & (3<<4)) >> 4)\
| ((PIND & (1<<1)) << 1)\
| ((PIND & (1<<0)) << 3)\
| ((PIND & (1<<4)) >> 0)\
| ((PINC & (1<<6)) >> 1)\
| ((PIND & (1<<7)) >> 1)\
| ((PINE & (1<<6)) << 1)\
)
#define setWriteDir() { DDRB |=  BMASK; DDRC |=  CMASK; DDRD |=  DMASK; DDRE |=  EMASK;  }
#define setReadDir()  { DDRB &= ~BMASK; DDRC &= ~CMASK; DDRD &= ~DMASK; DDRE &= ~EMASK;  }
#define write8(x)     { write_8(x); WR_STROBE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; dst = read_8(); RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }

#define PIN_LOW(p, b)        (p) &= ~(1<<(b))
#define PIN_HIGH(p, b)       (p) |= (1<<(b))
#define PIN_OUTPUT(p, b)     *(&p-1) |= (1<<(b))

#elif defined(__MK20DX128__) || defined(__MK20DX256__) // regular UNO shield on a Teensy 3.x
#warning regular UNO shield on a Teensy 3.x
#define RD_PORT GPIOD
#define RD_PIN  1
#define WR_PORT GPIOC
#define WR_PIN  0
#define CD_PORT GPIOB
#define CD_PIN  0
#define CS_PORT GPIOB
#define CS_PIN  1
#define RESET_PORT GPIOB
#define RESET_PIN  3

// configure macros for the data pins
#define AMASK ((1<<12)|(1<<13))
#define CMASK ((1<<3))
#define DMASK ((1<<0)|(1<<2)|(1<<3)|(1<<4)|(1<<7))

#define write_8(d) { \
   GPIOA_PCOR = AMASK; GPIOC_PCOR = CMASK; GPIOD_PCOR = DMASK; \
   GPIOA_PSOR = (((d) & (1<<3)) << 9) \
              | (((d) & (1<<4)) << 9); \
   GPIOC_PSOR = (((d) & (1<<1)) << 2); \
   GPIOD_PSOR = (((d) & (1<<0)) << 3) \
              | (((d) & (1<<2)) >> 2) \
              | (((d) & (1<<5)) << 2) \
              | (((d) & (1<<6)) >> 2) \
              | (((d) & (1<<7)) >> 5); \
  } 
#define read_8() (          (((GPIOD_PDIR & (1<<3)) >> 3) \
                             | ((GPIOC_PDIR & (1<<3)) >> 2) \
                             | ((GPIOD_PDIR & (1<<0)) << 2) \
                             | ((GPIOA_PDIR & (1<<12)) >> 9) \
                             | ((GPIOA_PDIR & (1<<13)) >> 9) \
                             | ((GPIOD_PDIR & (1<<7))  >> 2) \
                             | ((GPIOD_PDIR & (1<<4))  << 2) \
                             | ((GPIOD_PDIR & (1<<2))  << 5)))
#define setWriteDir() {GPIOA_PDDR |=  AMASK;GPIOC_PDDR |=  CMASK;GPIOD_PDDR |=  DMASK; }
#define setReadDir()  {GPIOA_PDDR &= ~AMASK;GPIOC_PDDR &= ~CMASK;GPIOD_PDDR &= ~DMASK; }

#define write8(x)     { write_8(x); WR_ACTIVE; WR_ACTIVE; WR_STROBE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; RD_ACTIVE; RD_ACTIVE; RD_ACTIVE; RD_ACTIVE; RD_ACTIVE; dst = read_8(); RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }

#define PASTE(x, y)   x ## y

#define PIN_LOW(port, pin)    PASTE(port, _PCOR) =  (1<<(pin))
#define PIN_HIGH(port, pin)   PASTE(port, _PSOR) =  (1<<(pin))
#define PIN_OUTPUT(port, pin) PASTE(port, _PDDR) |= (1<<(pin))

#elif defined(__STM32F1__) && defined(ARDUINO_STM_NUCLEO_F103RB) // Uno Shield on NUCLEO-F103
#warning Uno Shield on NUCLEO - F103 REGS
#define RD_PORT GPIOA
#define RD_PIN  0
#define WR_PORT GPIOA
#define WR_PIN  1
#define CD_PORT GPIOA
#define CD_PIN  4
#define CS_PORT GPIOB
#define CS_PIN  0
#define RESET_PORT GPIOC
#define RESET_PIN  1

// configure macros for the data pins
#define write_8(d) { \
   GPIOA->regs->BSRR = 0x0700 << 16; \
   GPIOB->regs->BSRR = 0x0438 << 16; \
   GPIOC->regs->BSRR = 0x0080 << 16; \
   GPIOA->regs->BSRR = (((d) & (1<<0)) << 9) \
               | (((d) & (1<<2)) << 8) \
               | (((d) & (1<<7)) << 1); \
   GPIOB->regs->BSRR = (((d) & (1<<3)) << 0) \
               | (((d) & (1<<4)) << 1) \
               | (((d) & (1<<5)) >> 1) \
               | (((d) & (1<<6)) << 4); \
   GPIOC->regs->BSRR = (((d) & (1<<1)) << 6); \
    }

#define read_8() (          (((GPIOA->regs->IDR & (1<<9)) >> 9) \
                             | ((GPIOC->regs->IDR & (1<<7)) >> 6) \
                             | ((GPIOA->regs->IDR & (1<<10)) >> 8) \
                             | ((GPIOB->regs->IDR & (1<<3)) >> 0) \
                             | ((GPIOB->regs->IDR & (1<<5)) >> 1) \
                             | ((GPIOB->regs->IDR & (1<<4)) << 1) \
                             | ((GPIOB->regs->IDR & (1<<10)) >> 4) \
                             | ((GPIOA->regs->IDR & (1<<8))  >> 1)))

// be wise to clear both MODER bits properly.
#define clearModebits()  { \
    GPIOA->regs->CRH &= ~0xFFF; GPIOB->regs->CRH &= ~0xF00; GPIOB->regs->CRL &= ~0xFFF000; GPIOC->regs->CRL &= ~0xF0000000; \
  }
#define setWriteDir() { \
    clearModebits(); \
    GPIOA->regs->CRH |=  0x333; GPIOB->regs->CRH |=  0x300; GPIOB->regs->CRL |=  0x333000; GPIOC->regs->CRL |=  0x30000000; \
  }
#define setReadDir()  { \
    clearModebits(); \
    GPIOA->regs->CRH |=  0x444; GPIOB->regs->CRH |=  0x400; GPIOB->regs->CRL |=  0x444000; GPIOC->regs->CRL |=  0x40000000; \
  }

#define write8(x)     { write_8(x); WR_ACTIVE; WR_STROBE; }
#define write16(x)    { uint8_t h = (x)>>8, l = x; write8(h); write8(l); }
#define READ_8(dst)   { RD_STROBE; RD_ACTIVE; RD_ACTIVE; RD_ACTIVE; RD_ACTIVE; RD_ACTIVE; dst = read_8(); RD_IDLE; }
#define READ_16(dst)  { uint8_t hi; READ_8(hi); READ_8(dst); dst |= (hi << 8); }

#define PIN_MODE4(reg, pin, mode) reg=(reg&~(0xF<<((pin)<<2)))|(mode<<((pin)<<2))

#define PIN_HIGH(port, pin)   (port)->regs->BSRR = (1<<(pin))
//#define PIN_LOW(port, pin)    (port)->regs->BSRR = (1<<((pin)+16))
#define PIN_LOW(port, pin)    (port)->regs->ODR &= ~(1<<(pin))
#define PIN_OUTPUT(port, pin) PIN_MODE4((port)->regs->CRL, pin, 0x3)   //50MHz push-pull only 0-7
#define PIN_INPUT(port, pin)  PIN_MODE4((port)->regs->CRL, pin, 0x4)   //digital input 

#else
#error MCU unsupported
#endif                          // regular UNO shields on Arduino boards

#endif                          //!defined(USE_SPECIAL) || defined (USE_SPECIAL_FAIL)

#define RD_ACTIVE  PIN_LOW(RD_PORT, RD_PIN)
#define RD_IDLE    PIN_HIGH(RD_PORT, RD_PIN)
#define RD_OUTPUT  PIN_OUTPUT(RD_PORT, RD_PIN)
#define WR_ACTIVE  PIN_LOW(WR_PORT, WR_PIN)
#define WR_IDLE    PIN_HIGH(WR_PORT, WR_PIN)
#define WR_OUTPUT  PIN_OUTPUT(WR_PORT, WR_PIN)
#define CD_COMMAND PIN_LOW(CD_PORT, CD_PIN)
#define CD_DATA    PIN_HIGH(CD_PORT, CD_PIN)
#define CD_OUTPUT  PIN_OUTPUT(CD_PORT, CD_PIN)
#define CS_ACTIVE  PIN_LOW(CS_PORT, CS_PIN)
#define CS_IDLE    PIN_HIGH(CS_PORT, CS_PIN)
#define CS_OUTPUT  PIN_OUTPUT(CS_PORT, CS_PIN)
#define RESET_ACTIVE  PIN_LOW(RESET_PORT, RESET_PIN)
#define RESET_IDLE    PIN_HIGH(RESET_PORT, RESET_PIN)
#define RESET_OUTPUT  PIN_OUTPUT(RESET_PORT, RESET_PIN)

// General macros.   IOCLR registers are 1 cycle when optimised.
#define WR_STROBE { WR_ACTIVE; WR_IDLE; }       //PWLW=TWRL=50ns
#define RD_STROBE RD_IDLE, RD_ACTIVE, RD_ACTIVE, RD_ACTIVE      //PWLR=TRDL=150ns, tDDR=100ns

#if defined(TEENSYDUINO)
#warning special pinMode() initialisation for TEENSYDUINO
#define CTL_INIT()   { \
                       for (int i = 2; i <= 9; i++) pinMode(i, OUTPUT); \
                       for (int i = A0; i <= A4; i++) pinMode(i, OUTPUT); \
                       RD_OUTPUT; WR_OUTPUT; CD_OUTPUT; CS_OUTPUT; RESET_OUTPUT; }
#else
#define CTL_INIT()   { RD_OUTPUT; WR_OUTPUT; CD_OUTPUT; CS_OUTPUT; RESET_OUTPUT; }
#endif
#define WriteCmd(x)  { CD_COMMAND; write16(x); }
#define WriteData(x) { CD_DATA; write16(x); }


#endif
